//------------------------------------------------------------------------------
//  Pink Noise Object for Pd
//
//  pink~.c
//
//  Generates pink noise
//
//  Based on the coefficients by Paul Kellet
//  http://www.musicdsp.org/files/pink.txt
//
//  Created by Cooper Baker on 11/1/14.
//  Copyright (c) 2014 Cooper Baker. All rights reserved.
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
// headers
//------------------------------------------------------------------------------

// main header for pd
#include "m_pd.h"

// standard library for random number generator
#include <stdlib.h>

// time header to seed random numbers
#include <time.h>

// utility header for Pd Spectral Toolkit project
#include "utility.h"

// disable compiler warnings on windows
#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif


//------------------------------------------------------------------------------
// pink_class - pointer to this object's definition
//------------------------------------------------------------------------------
static t_class* pink_class;


//------------------------------------------------------------------------------
// pink - data structure holding this object's data
//------------------------------------------------------------------------------
typedef struct pink
{
    // this object - must always be first variable in struct
    t_object object;

    // coefficients
    double b0, b1, b2, b3, b4, b5, b6;

    // normalizes rand() output
    double rand_norm;
    
} t_pink;


//------------------------------------------------------------------------------
// function prototypes
//------------------------------------------------------------------------------
static t_int* pink_perform     ( t_int* io );
static void   pink_dsp         ( t_pink* object, t_signal **sig );
static void*  pink_new         ( void );
void          pink_tilde_setup ( void );


//------------------------------------------------------------------------------
// pink_perform - the signal processing function of this object
//
// Adapted from Paul Kellet's technique at:
// http://www.firstpr.com.au/dsp/pink-noise/
// http://musicdsp.org/files/pink.txt
//------------------------------------------------------------------------------
static t_int* pink_perform( t_int* io )
{
    // store variables from dsp input/output array
    t_float* out    = ( t_float* )( io[ 1 ] );
    t_int    frames = ( t_int    )( io[ 2 ] );
    t_pink*  object = ( t_pink*  )( io[ 3 ] );

    // copy pointers to values from object struct
    double* b0 = &object->b0;
    double* b1 = &object->b1;
    double* b2 = &object->b2;
    double* b3 = &object->b3;
    double* b4 = &object->b4;
    double* b5 = &object->b5;
    double* b6 = &object->b6;
    
    // copy rand normalization value
    double rand_norm = object->rand_norm;
  
    // noise variables
    double white;
    double pink;

    // signal vector iterator variable
    t_int n = -1;
    
    // the dsp loop
    while( ++n < frames )
    {
        white = rand() * rand_norm - 1.0;
        
        *b0 =  0.99886 * *b0 + white * 0.0555179;
        *b1 =  0.99332 * *b1 + white * 0.0750759;
        *b2 =  0.96900 * *b2 + white * 0.1538520;
        *b3 =  0.86650 * *b3 + white * 0.3104856;
        *b4 =  0.55000 * *b4 + white * 0.5329522;
        *b5 = -0.7616  * *b5 - white * 0.0168980;

        pink = *b0 + *b1 + *b2 + *b3 + *b4 + *b5 + *b6 + white * 0.5362;

        *b6 = white * 0.115926;
        
        out[ n ] = pink;
    }
    
    // return the dsp input/output array address plus one more than its size
    // to provide a pointer to the next perform function in pd's call list
    return &( io[ 4 ] );
}


//------------------------------------------------------------------------------
// pink_dsp - installs this object's dsp function in pd's callback list
//------------------------------------------------------------------------------
static void pink_dsp( t_pink* object, t_signal **sig )
{
    // dsp_add arguments
    //--------------------------------------------------------------------------
    // perform routine
    // number of passed parameters
    // outlet sample vector
    // sample frames to process (vector size)
    // object pointer
    dsp_add( pink_perform, 3, sig[ 0 ]->s_vec, sig[ 0 ]->s_n, object );
}


//------------------------------------------------------------------------------
// pink_new - instantiates a copy of this object in pd
//------------------------------------------------------------------------------
static void* pink_new( void )
{
    // create a pointer to this object
    t_pink* object = ( t_pink* )pd_new( pink_class );
    
    // create a new signal outlet for this object
    outlet_new( &object->object, gensym( "signal" ) );
    
    // set up coefficients
    object->b0 = 0;
    object->b1 = 0;
    object->b2 = 0;
    object->b0 = 0;
    object->b1 = 0;
    object->b2 = 0;
    object->b6 = 0;
    
    // calculate rand normalization value
    object->rand_norm = ( 2.0 / RAND_MAX );
    
    // seed the random number generator
    time_t t;
    srand( ( unsigned )time( &t ) );
    
    return object;
}


//------------------------------------------------------------------------------
// pink_tilde_setup - describes the attributes of this object to pd so it may be properly instantiated
// (must always be named with _tilde replacing ~ in the object name)
//------------------------------------------------------------------------------
void pink_tilde_setup( void )
{
    // creates an instance of this object and describes it to pd
    pink_class = class_new( gensym( "pink~" ), ( t_newmethod )pink_new, 0, sizeof( t_pink ), CLASS_NOINLET, 0 );
    
    // installs pink_dsp so that it will be called when dsp is turned on
    class_addmethod( pink_class, ( t_method )pink_dsp, gensym( "dsp" ), 0 );
    
    // announce this object in the pd console
    Announce( "pink~: pink noise generator - v1.0" );
}


//------------------------------------------------------------------------------
// EOF
//------------------------------------------------------------------------------
