//------------------------------------------------------------------------------
//  Pink Noise Object for Pd
//
//  pink~.c
//
//  Generates pink noise
//
//  Based on the technique by Larry Trammell and code by David Lowenfels
//  http://musicdsp.org/archive.php?classid=1#220
//
//  Created by Cooper Baker on 11/1/14.
//  Copyright (c) 2014 Cooper Baker. All rights reserved.
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
// headers
//------------------------------------------------------------------------------

// main header for pd
#include "m_pd.h"

// standard library for random number generator
#include <stdlib.h>

// time header to seed random numbers
#include <time.h>

// utility header for Pd Spectral Toolkit project
#include "utility.h"

// disable compiler warnings on windows
#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif


//------------------------------------------------------------------------------
// pink_class - pointer to this object's definition
//------------------------------------------------------------------------------
static t_class* pink_class;


//------------------------------------------------------------------------------
// pink - data structure holding this object's data
//------------------------------------------------------------------------------
typedef struct pink
{
    // this object - must always be first variable in struct
    t_object object;

    // coefficients
    double a0, a1, a2;
    double p0, p1, p2;
    
    // filter states
    double s0, s1, s2;
    
    // 2 over rand max constant
    double rm;
    
    // coefficient offset
    double of;
    
} t_pink;


//------------------------------------------------------------------------------
// function prototypes
//------------------------------------------------------------------------------
static t_int* pink_perform     ( t_int* io );
static void   pink_dsp         ( t_pink* object, t_signal **sig );
static void*  pink_new         ( void );
void          pink_tilde_setup ( void );


//------------------------------------------------------------------------------
// pink_perform - the signal processing function of this object
//------------------------------------------------------------------------------
static t_int* pink_perform( t_int* io )
{
    // store variables from dsp input/output array
    t_float* out    = ( t_float* )( io[ 1 ] );
    t_int    frames = ( t_int    )( io[ 2 ] );
    t_pink*  o      = ( t_pink*  )( io[ 3 ] );

    // copy values from object struct
    double a0 = o->a0;
    double a1 = o->a1;
    double a2 = o->a2;
    double p0 = o->p0;
    double p1 = o->p1;
    double p2 = o->p2;
    double rm = o->rm;
    double of = o->of;
    
    // random number variable
    double rnd;

    // signal vector iterator variable
    t_int n = -1;
    
    // the dsp loop
    while( ++n < frames )
    {
        rnd = ( float )rand();
        o->s0 = p0 * ( o->s0 - rnd ) + rnd;
        
        rnd = ( float )rand();
        o->s1 = p1 * ( o->s1 - rnd ) + rnd;
        
        rnd = ( float )rand();
        o->s2 = p2 * ( o->s2 - rnd ) + rnd;
        
        out[ n ] = ( a0 * o->s0 + a1 * o->s1 + a2 * o->s2 ) * rm - of;
    }
    
    // return the dsp input/output array address plus one more than its size
    // to provide a pointer to the next perform function in pd's call list
    return &( io[ 4 ] );
}


//------------------------------------------------------------------------------
// pink_dsp - installs this object's dsp function in pd's callback list
//------------------------------------------------------------------------------
static void pink_dsp( t_pink* object, t_signal **sig )
{
    // dsp_add arguments
    //--------------------------------------------------------------------------
    // perform routine
    // number of passed parameters
    // outlet sample vector
    // sample frames to process (vector size)
    // object pointer
    dsp_add( pink_perform, 3, sig[ 0 ]->s_vec, sig[ 0 ]->s_n, object );
}


//------------------------------------------------------------------------------
// pink_new - instantiates a copy of this object in pd
//------------------------------------------------------------------------------
static void* pink_new( void )
{
    // create a pointer to this object
    t_pink* object = ( t_pink* )pd_new( pink_class );
    
    // create a new signal outlet for this object
    outlet_new( &object->object, gensym( "signal" ) );
    
    // set up coefficients
    object->a0 = 0.02109238;
    object->a1 = 0.07113478;
    object->a2 = 0.68873558;
    object->p0 = 0.3190;
    object->p1 = 0.7756;
    object->p2 = 0.9613;
    
    // initialize filter states
    object->s0 = 0;
    object->s1 = 0;
    object->s2 = 0;
    
    // calculate 2 over rand max
    object->rm = 2.0 / ( t_float )RAND_MAX;
    
    // calculate offset
    object->of = object->a0 + object->a1 + object->a2;
    
    // seed the random number generator
    time_t t;
    srand( ( unsigned )time( &t ) );
    
    return object;
}


//------------------------------------------------------------------------------
// pink_tilde_setup - describes the attributes of this object to pd so it may be properly instantiated
// (must always be named with _tilde replacing ~ in the object name)
//------------------------------------------------------------------------------
void pink_tilde_setup( void )
{
    // creates an instance of this object and describes it to pd
    pink_class = class_new( gensym( "pink~" ), ( t_newmethod )pink_new, 0, sizeof( t_pink ), CLASS_NOINLET, 0 );
    
    // installs pink_dsp so that it will be called when dsp is turned on
    class_addmethod( pink_class, ( t_method )pink_dsp, gensym( "dsp" ), 0 );
    
    // announce this object in the pd console
    Announce( "pink~: pink noise generator - v1.0" );
}


//------------------------------------------------------------------------------
// EOF
//------------------------------------------------------------------------------
